if get_material("system/SkyBox") == nil then
    local sb = Material "system/SkyBox"
    sb:setFogOverride(0,0,true)
    sb:setLighting(0,0,false)
    sb:setPolygonModeOverrideable(0,0,false)
    sb:createTextureUnitState(0,0)
    sb:setAddressMode(0,0,0,"CLAMP")
    mat:setTextureFiltering(0,0,0,"LINEAR","LINEAR","LINEAR")
    mat:setCubicTextureName(0,0,0,"system/skybox_side.dds", "system/skybox_side.dds", 
                                  "system/skybox_side.dds", "system/skybox_side.dds", 
                                  "system/skybox_up.dds", "system/skybox_down.dds")
end
sm:setSkyBox(true,"system/SkyBox",100,true,Q_UP)

function format_time (secs)
    return string.format("%02d:%02d:%02d", math.mod(math.floor(secs/60/60),24),
                                           math.mod(math.floor(secs/60),60),
                                           math.mod(secs,60))
end

local timer = Timer()

local function maybe_commit (self)
    if not self.c.autoUpdate then return end

    local reset_sun_pos = false
    local reset_fog = false

    for k,v in pairs(self.p) do
        if self.c[k] ~= v then
            self.c[k] = v

            if k == "sunDirection" then
                sun.direction = Vector3(v[1],v[2],v[3]).normalised
            elseif k == "sunDiffuse" then
                sun:setDiffuseColour(v[1],v[2],v[3])
            elseif k == "sunSpecular" then
                sun:setSpecularColour(v[1],v[2],v[3])
            elseif k == "ambient" then
                sm:setAmbientLight(v[1],v[2],v[3])
            elseif k == "secondsSinceMidnight" then
                reset_sun_pos = true
            elseif k == "latitude" then
                reset_sun_pos = true
            elseif k == "season" then
                reset_sun_pos = true
            elseif k == "earthTilt" then
                reset_sun_pos = true
            elseif k == "clockRate" then
            elseif k == "clockTicking" then
                if v then timer:reset() end
            elseif k == "fogColour" then
                reset_fog = true
            elseif k == "fogStrength" then
                reset_fog = true
            else
                error("Unexpected: "..k)
            end
        end
    end

    if reset_fog then
        sm:setFog("EXP2", self.fogColour[1],self.fogColour[2],self.fogColour[3], self.fogStrength, 0,0)
    end

    if reset_sun_pos then
        local secs = self.secondsSinceMidnight
        local time = secs / 60 / 60 / 24 * 360
        -- account for the fact that the sun is on the other side of the planet in summer
        local adjusted_time = math.mod(time + self.season, 360)
        local sun_direction =  - Quat(self.season,V_NORTH)*
                                 Quat(self.earthTilt,V_EAST)*
                                 Quat(adjusted_time,V_SOUTH)*
                                 Quat(self.latitude,V_WEST) * V_UP
        local moon_direction = (- sun_direction + 0.5*V_DOWN).normalised
        local found = false
        local current_sky
        local next_sky
        for _,sky in ipairs(sky_cycle) do
            current_sky = next_sky
            next_sky = sky
            if next_sky.time > secs then
                found = true
                break
            end
        end
        local slider
        if not found then
            current_sky = sky_cycle[#sky_cycle]
            next_sky = sky_cycle[1]
            slider = (secs - current_sky.time) / (next_sky.time*24*60*60 - current_sky.time)
        else
            slider = (secs - current_sky.time) / (next_sky.time - current_sky.time)
        end
        local current_diff = current_sky.sunDiffuse or {0,0,0}
        local next_diff =  next_sky.sunDiffuse or {0,0,0}
        sun:setDiffuseColour(lerp(slider, current_diff[1], next_diff[1]),
                             lerp(slider, current_diff[2], next_diff[2]),
                             lerp(slider, current_diff[3], next_diff[3]))
        local current_spec = current_sky.sunSpecular or current_diff
        local next_spec =  next_sky.sunSpecular or  next_diff
        sun:setSpecularColour(lerp(slider, current_spec[1], next_spec[1]),
                              lerp(slider, current_spec[2], next_spec[2]),
                              lerp(slider, current_spec[3], next_spec[3]))
        local current_ambient = current_sky.ambient or {0,0,0}
        local next_ambient =  next_sky.ambient or {0,0,0}
        sm:setAmbientLight(lerp(slider, current_ambient[1], next_ambient[1]),
                           lerp(slider, current_ambient[2], next_ambient[2]),
                           lerp(slider, current_ambient[3], next_ambient[3]))
        local current_fog_colour = current_sky.fogColour or {0,0,0}
        local next_fog_colour =  next_sky.fogColour or {0,0,0}
        local current_fog_strength = current_sky.fogStrength or 0
        local next_fog_strength =  next_sky.fogStrength or 0
        sm:setFog("EXP2", lerp(slider, current_fog_colour[1], next_fog_colour[1]),
                          lerp(slider, current_fog_colour[2], next_fog_colour[2]),
                          lerp(slider, current_fog_colour[3], next_fog_colour[3]),
                          lerp(slider, current_fog_strength, next_fog_strength),
                          0, 0)

        if current_sky.moonLight then
            sun.direction = moon_direction
        else
            sun.direction = sun_direction
        end
    end
    
end

local function change (self, k, v)
    if k == "autoUpdate" then
        ensure_one_of(v,{false,true})
        self.c[k] = v
    elseif k == "sunDirection" then
        ensure_array(v,3)
        self.p[k] = v
    elseif k == "sunDiffuse" then
        ensure_array(v,3)
        ensure_range(v[1],0,1)
        ensure_range(v[2],0,1)
        ensure_range(v[3],0,1)
        self.p[k] = v
    elseif k == "sunSpecular" then
        ensure_array(v,3)
        ensure_range(v[1],0,1)
        ensure_range(v[2],0,1)
        ensure_range(v[3],0,1)
        self.p[k] = v
    elseif k == "ambient" then
        ensure_array(v,3)
        ensure_range(v[1],0,1)
        ensure_range(v[2],0,1)
        ensure_range(v[3],0,1)
        self.p[k] = v
    elseif k == "secondsSinceMidnight" then
        ensure_number(v)
        self.p[k] = v
    elseif k == "latitude" then
        ensure_range(v,-90,90)
        self.p[k] = v
    elseif k == "season" then
        ensure_range(v,0,360)
        self.p[k] = v
    elseif k == "earthTilt" then
        ensure_range(v,-90,90)
        self.p[k] = v
    elseif k == "clockRate" then
        ensure_range(v,-50000,50000)
        self.p[k] = v
    elseif k == "clockTicking" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "fogColour" then
        ensure_array(v,3)
        ensure_range(v[1],0,1)
        ensure_range(v[2],0,1)
        ensure_range(v[3],0,1)
        self.p[k] = v
    elseif k == "fogStrength" then
        ensure_range(v,0,1)
        self.p[k] = v

    else
        error("Unrecognised env setting: "..tostring(k))
    end

    maybe_commit(self)
end

local function frameCallback()

    local clock_rate = env.clockRate

    if env.clockTicking then

        local elapsed = timer.s 
        timer:reset()

        elapsed = elapsed * env.clockRate

        env.secondsSinceMidnight = math.mod(env.secondsSinceMidnight + elapsed, 24*60*60)

    end
end

sky_cycle = {
    { time=( 0*60*60), ambient={0.10, 0.10, 0.15}, fogColour={0.00,0.00,0.00}, fogStrength=0.0025, moonLight=true,  sunDiffuse={0.40, 0.50, 0.60} },
    { time=( 3*60*60), ambient={0.10, 0.10, 0.15}, fogColour={0.00,0.00,0.00}, fogStrength=0.0015, moonLight=true,  sunDiffuse={0.30, 0.40, 0.50} },
    { time=( 4*60*60), ambient={0.10, 0.10, 0.10}, fogColour={0.00,0.00,0.00}, fogStrength=0.0015, moonLight=false, },
    { time=( 6*60*60), ambient={0.12, 0.12, 0.12}, fogColour={0.70,0.50,0.40}, fogStrength=0.0015, moonLight=false, sunDiffuse={1.00, 0.50, 0.30} },
    { time=( 8*60*60), ambient={0.13, 0.13, 0.13}, fogColour={0.70,0.50,0.40}, fogStrength=0.0015, moonLight=false, sunDiffuse={1.00, 1.00, 0.50} },
    { time=(10*60*60), ambient={0.14, 0.14, 0.14}, fogColour={0.70,0.50,0.40}, fogStrength=0.0015, moonLight=false, sunDiffuse={1.00, 1.00, 0.80} },
    { time=(12*60*60), ambient={0.14, 0.14, 0.14}, fogColour={0.70,0.50,0.40}, fogStrength=0.0010, moonLight=false, sunDiffuse={1.00, 1.00, 0.80} },
    { time=(17*60*60), ambient={0.14, 0.14, 0.14}, fogColour={0.70,0.50,0.40}, fogStrength=0.0015, moonLight=false, sunDiffuse={1.00, 1.00, 0.80} },
    { time=(18*60*60), ambient={0.12, 0.12, 0.12}, fogColour={0.00,0.00,0.00}, fogStrength=0.0015, moonLight=false, sunDiffuse={1.00, 0.70, 0.20}, sunSpecular={1.00, 0.50, 0.50} },
    { time=(20*60*60), ambient={0.10, 0.10, 0.10}, fogColour={0.00,0.00,0.00}, fogStrength=0.0015, moonLight=true, },
    { time=(21*60*60), ambient={0.10, 0.10, 0.15}, fogColour={0.00,0.00,0.00}, fogStrength=0.0025, moonLight=true,  sunDiffuse={0.40, 0.50, 0.60} },
}


if not env then
    env = {
        -- current values
        c = {
            autoUpdate = false;
        },

        -- proposed values
        p = {
            sunDirection = {0,0,-1};
            sunDiffuse = {1,1,1};
            sunSpecular = {1,1,1};
            ambient = {.2, .2, .2};
            latitude = 55; 
            secondsSinceMidnight = 12*60*60; --midday
            season = 180; -- summer, stored as angle in degrees
            earthTilt = 30;
            clockRate = 600;
            clockTicking = true;
            fogColour = {.7137,.5333,.4}; -- colour
            fogStrength = 0.0015; -- strength
            
        }
    }

    sun = sm:createLight("Sun")
    sun.type = "DIRECTIONAL"


else

    main.frameCallbacks:removeByName("Environment")
    
end

main.frameCallbacks:insert("Environment", frameCallback)

setmetatable(env, {
    __index = function (self, k)
        local v = self.c[k] ;
        if v == nil then
            error("No such setting: \""..k.."\"",2)
        end
        return v
    end,
    __newindex = function (self, k, v) change(self,k,v) end
})

env.autoUpdate = true

ui:bind("F2",function() env.clockTicking = not env.clockTicking end)
